#include "gen_classes.h"

// IPv6 and 6lowpan codec implementation
//
// Please read the Codec Generation section of the T3DevKit documentation in
// order to understand how this CoDec is designed.
//
// --->  http://t3devkit.gforge.inria.fr/doc/userref/
//

namespace t3devlib {
namespace gen {
	
	///////////////////////////////////////////////////////////////////////////////////////////
	// IPv6 extension headers

	void ExtensionHeaderList::SetHypNextHeader (int nh)
	{
		switch (nh)
		{
		case 0:
			ExtensionHeader::SetHypChosenId (ExtensionHeader::id_hopByHopHeader);
			break;
		case 6:  // tcp
		case 17: // udp
		case 58: // udp
		case 59: // no next header
		case 77: // debug value
			SetHypSize (-2);
			break;
			
		case 43:
			ExtensionHeader::SetHypChosenId (ExtensionHeader::id_routingHeader);
			break;

		case 44:
			ExtensionHeader::SetHypChosenId (ExtensionHeader::id_fragmentHeader);
			break;
			
		case 60:
			ExtensionHeader::SetHypChosenId (ExtensionHeader::id_destinationOptionHeader);
			break;
/*			
		default:
			ExtensionHeader::SetHypChosenId (ExtensionHeader::id_otherHeader);
			break;
*/
		}
	}
	
	// Template for decoding an extension header w/o a length field
	template<class T>
	void ExtensionHeader_CommonPreDecode (T& msg, Buffer& buffer) throw (DecodeError)
	{
		Unsigned nh(8);
		int position = buffer.GetPosition();

		// next header
		nh.Decode (buffer);
		Variable* parent = msg.GetParent();
		if (parent && dynamic_cast<ExtensionHeaderList*> (parent)) {
			static_cast<ExtensionHeaderList*> (parent) -> SetHypNextHeader (nh.GetValue());
		} else {
			cerr << "ExtensionHeader_CommonPreDecode: cannot access parent ExtensionHeaderList" << endl;
		}
		
		buffer.SetPosition(position);
	}

	// Template for decoding an extension header with a length field
	template<class T>
	void ExtensionHeader_CommonPreDecode_WithLength (T& msg, Buffer& buffer) throw (DecodeError)
	{
		int position = buffer.GetPosition();
		Unsigned u8(8);

		ExtensionHeader_CommonPreDecode (msg, buffer);
		
		// next header
		u8.Decode (buffer);

		// length
		u8.Decode (buffer);
		T::SetHypLength (u8.GetValue() * 64 + 64);

		buffer.SetPosition(position);
	}
	
	void HopByHopHeader::PreDecode (Buffer& buffer) throw (DecodeError)
	{
		ExtensionHeader_CommonPreDecode (*this, buffer);
	}
	
	void RoutingHeader::PreDecode (Buffer& buffer) throw (DecodeError)
	{
		ExtensionHeader_CommonPreDecode (*this, buffer);
	}

	void FragmentHeader::PreDecode (Buffer& buffer) throw (DecodeError)
	{
		ExtensionHeader_CommonPreDecode (*this, buffer);
	}
	
	void DestOptionHeader::PreDecode (Buffer& buffer) throw (DecodeError)
	{
		ExtensionHeader_CommonPreDecode (*this, buffer);
	}
/*	
	void OtherIpv6ExtHdr::PreDecode (Buffer& buffer) throw (DecodeError)
	{
		ExtensionHeader_CommonPreDecode (*this, buffer);
	}
*/
	
	///////////////////////////////////////////////////////////////////////////////////////////
	// IPv6 messages
	
	// Template for decoding IPv6 messages
	// -> decode extension headers
	template<class T>
	void IPv6_CommonPreDecodeField (T& msg, int id, Buffer& buffer) throw (DecodeError)
	{
		if (id == T::id_extHdrList)
		{
			int nh = msg.Get_ipv6Hdr().Get_nextHeader().GetValue();
			ExtensionHeaderList& ehl = msg.Get_extHdrList();

			ehl.SetHypNextHeader (nh);
			if (ehl.GetHypSize() == -2)
			{
				ehl.SetHypSize (-1);
				T::SetHypFieldIsPresent (T::id_extHdrList, 0);
			} else {
				T::SetHypFieldIsPresent (T::id_extHdrList, 1);
			}
		}
	}

	void GeneralIpv6::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		IPv6_CommonPreDecodeField (*this, id, buffer);

		if (id == id_data) {
			SetHypFieldIsPresent (id_data, buffer.GetBitsLeft() ? 1 : 0);
		}
	}

	void UdpPacket::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		IPv6_CommonPreDecodeField (*this, id, buffer);

		if (id == id_data) {
			SetHypFieldIsPresent (id_data, buffer.GetBitsLeft() ? 1 : 0);
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////
	// ICMPv6 messages

	// Template for decoding ICMPv6 messages
	// -> check if icmp_code field is valid
	// -> check if trailing bits are present to decide if we decode the optional field (if given)
	template<class T>
	void ICMPv6_CommonPreDecodeField (T& msg, int icmp_code, int id, Buffer& buffer, int id_optional = -1) throw (DecodeError)
	{
		IPv6_CommonPreDecodeField (msg, id, buffer);
			
		if (id == T::id_icmpCode)
		{
			if (msg.Get_icmpType().GetValue() != icmp_code) {
				DecodeError ex(&msg);
				ex.Msg() << "Invalid ICMPv6 type field: " << msg.Get_icmpType() << endl;
				throw ex;
			}
		} else if (id == id_optional) {
			T::SetHypFieldIsPresent (id, buffer.GetBitsLeft() ? 1 : 0);
		}
	}


	void ParameterProblem::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		ICMPv6_CommonPreDecodeField (*this, 4, id, buffer, id_data);
	}

	void EchoRequest::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		ICMPv6_CommonPreDecodeField (*this, 128, id, buffer, id_data);
	}

	void EchoReply::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		ICMPv6_CommonPreDecodeField (*this, 129, id, buffer, id_data);
	}
	
	void TimeExceeded::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		ICMPv6_CommonPreDecodeField (*this, 3, id, buffer, id_data);
	}

	void Redirect::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		ICMPv6_CommonPreDecodeField (*this, 137, id, buffer, id_redirectOptions);
	}

	void DestinationUnreachable::PreDecodeField (int id, Buffer& buffer) throw (DecodeError)
	{
		ICMPv6_CommonPreDecodeField (*this, 1, id, buffer, id_data);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////
	// IPv6 Packet

	// return true if the packet can be decoded successfully
	template <class T>
	bool Ipv6Packet::TryDecode (Buffer& buffer)
	{
		bool return_value;
		int position = buffer.GetPosition();

		try {
			T val;
			val.Decode (buffer);

			return_value = (buffer.GetBitsLeft() == 0);
		}
		catch (DecodeError& ex) {
			// failed !!!
			return_value = false;
		}
		
		buffer.SetPosition (position);

		return return_value;
	}
	
	void Ipv6Packet::PreDecode (Buffer& buffer) throw (DecodeError)
	{
		// brute force attack :-)
		// 
		if (TryDecode <DestinationUnreachable> (buffer)) {
			SetHypChosenId (id_destinationUnreachable);
		} else if (TryDecode <PacketTooBig> (buffer)) {
			SetHypChosenId (id_packetTooBig);
		} else if (TryDecode <TimeExceeded> (buffer)) {
			SetHypChosenId (id_timeExceeded);
		} else if (TryDecode <ParameterProblem> (buffer)) {
			SetHypChosenId (id_parameterProblem);
		} else if (TryDecode <EchoRequest> (buffer)) {
			SetHypChosenId (id_echoRequest);
		} else if (TryDecode <EchoReply> (buffer)) {
			SetHypChosenId (id_echoReply);
		} else if (TryDecode <RouterAdvertisement> (buffer)) {
			SetHypChosenId (id_routerAdvert);
		} else if (TryDecode <RouterSolicitation> (buffer)) {
			SetHypChosenId (id_routerSolicitation);
		} else if (TryDecode <NeighborSolicitation> (buffer)) {
			SetHypChosenId (id_nbrSolicitation);
		} else if (TryDecode <NeighborAdvertisement> (buffer)) {
			SetHypChosenId (id_nbrAdvert);
		} else if (TryDecode <Redirect> (buffer)) {
			SetHypChosenId (id_redirect);
		} else if (TryDecode <RouterRenumbering> (buffer)) {
			SetHypChosenId (id_routerRenumbering);
		} else if (TryDecode <HomeAgentAddressDiscoveryRequest> (buffer)) {
			SetHypChosenId (id_homeAgentAddressDiscoveryRequest);
		} else if (TryDecode <HomeAgentAddressDiscoveryReply> (buffer)) {
			SetHypChosenId (id_homeAgentAddressDiscoveryReply);
		} else if (TryDecode <MobilePrefixSolicitation> (buffer)) {
			SetHypChosenId (id_mobilePrefixSolicitation);
		} else if (TryDecode <MobilePrefixAdvertisement> (buffer)) {
			SetHypChosenId (id_mobilePrefixAdvertisement);
		} else if (TryDecode <MipRouterAdvertisement> (buffer)) {
			SetHypChosenId (id_mipRouterAdvertisement);
		} else if (TryDecode <OtherIcmpv6Hdr> (buffer)) {
			SetHypChosenId (id_otherHeader);
		} else if (TryDecode <UdpPacket> (buffer)) {
			SetHypChosenId (id_udpPacket);
		} else {
			SetHypChosenId (id_generalIpv6);
		}
	}
}}
    
