
// 6lowpan System Adapter implementation
//
// Please read the "Implementing Adapters" section of the T3DevKit
// documentation in order to understand how this SA is designed.
//
// --->  http://t3devkit.gforge.inria.fr/doc/userref/


#include <sys/types.h>
#include <cstdio>
#include <boost/regex.hpp>
#include <fstream>

#include "6lowpan-port.h"
#include "802.15.4-framer.h"


#define max(a,b)	( (a)>(b) ? (a) : (b) )

using namespace std;
using namespace t3devlib;

namespace boostai = boost::asio::ip;

static uint16_t my_htons (uint16_t in)
{
	uint16_t out;
	((uint8_t*) &out) [0] = in >> 8;
	((uint8_t*) &out) [1] = in & 0xff;
	return out;
}
//////////////////////////////////////////////////////////////////////////////////////////////////

void SixLowpanPort::StartReceive()
{
	mSocket.async_receive (
		boost::asio::buffer (mBuffer),
		boost::bind (
			&SixLowpanPort::HandleReceive, this,
			boost::asio::placeholders::error,
           		 boost::asio::placeholders::bytes_transferred
	));
}
void SixLowpanPort::HandleReceive (const boost::system::error_code& error, std::size_t len)
{
	bool result;

	switch (error.value())
	{
	case 0:
		if (len) {
			int	hw_src_len, hw_dst_len, payload_len;
			uint8_t	hw_src[8],  hw_dst[8];
			const void* payload;
			uint16_t pan_id;
			uint8_t snd_buff[2048];
			uint8_t* p;

			if (decode_ieee_802_15_4_frame (mBuffer.data(), len,
				hw_src, hw_src_len,
				hw_dst, hw_dst_len,
				payload, payload_len,
				pan_id))
			{
				if (mPanId && (mPanId != pan_id))
				{
					cerr << "SixLowpanPort: ignoring received frame (pan id mismatch)" << endl;
				} else {
					p = snd_buff;

					p[0] = hw_src_len;
					p[1] = hw_dst_len;
					p+=2;

					memcpy (p, hw_src, hw_src_len);
					p+=hw_src_len;

					memcpy (p, hw_dst, hw_dst_len);
					p+=hw_dst_len;

					memcpy (p, payload, payload_len);
					p+=payload_len;

					EnqueueMsg (*mMappedPort, MappedBitstring (snd_buff, (p-snd_buff)*8));
				}
			} else {
				cerr << "SixLowpanPort: receive: ignored frame (malformed)" << endl;
			}

		}
		StartReceive();
		return;
	
	case boost::system::errc::operation_canceled:
		// socket close
		result = 1;
		break;

	default:
		// receive error
		cerr << "SixLowpanPort: receive error: " << error.message() << endl;
		result = 0;
	}
	boost::mutex::scoped_lock l (mMutex);
	mReceiveResult = result;
	mCondition.notify_one();
}

//////////////////////////////////////////////////////////////////////////////////////////////////


SixLowpanPort::SixLowpanPort (PortId& id) :
	Port (id),
 	mSocket (boost_asio_io_service()),
	mMappedPort (NULL)
{
}

SixLowpanPort::~SixLowpanPort()
{
	if (mMappedPort !=  NULL)
		Unmap (*mMappedPort);
}

bool SixLowpanPort::Map (const PortId& port_id)
{
	{
		boost::mutex::scoped_lock lock (mMutex);

		// Allow only one connection
		if (mMappedPort != NULL) {
			cerr << "SixLowpanPort: cannot map to more than one component" << endl;
			return false;
		}

		// Remember the port id
		mMappedPort = &port_id;

		// Reset the configured status
		mConfigured = false;
	}

	return true;
}

bool SixLowpanPort::Unmap (const PortId& port_id)
{
	boost::mutex::scoped_lock lock (mMutex);

	if (mMappedPort)
	{
		if (mConfigured) 
		{
			// Stop listening
			mSocket.close();

			// Wait for the return value
			if (mReceiveResult < 0)
				mCondition.wait (lock);
		} else {
			// nothing configured
			//  -> just return success
			mReceiveResult = 1;
		}

		// Unmap
		mMappedPort = NULL;

		return mReceiveResult;
	} else {
		return false;
	}
}

bool SixLowpanPort::Send (const ComponentId& from,const Bitstring& msg)
{
	assert (mMappedPort);

	if (!mConfigured)
	{
		// the port is not yet configured
		// -> assume this is a configuration frame
		
		const uint8_t* p = msg.GetValueBin();
		int len = msg.GetLength()/8;
		
		if (len < 4) {
			cerr << "SixLowpanPort::Send() error: configuration frame is too short" << endl;
			return false;
		}

		// configure PAN ID
		mPanId = *(uint16_t*)p;
		p += 2;

		// remote IP of the phy adapter
		string remote_ip ((char*)p, len-4);
		p += len-4;

		// remote port of the phy adapter
		uint16_t remote_port = my_htons (*(uint16_t*)p);

		// connect the socket
		try
		{
			mSocket.connect (boostai::udp::endpoint (
				boostai::address_v4::from_string (remote_ip),
				remote_port
			));
		} catch (boost::system::system_error& error) {
			cerr << "SixLowpanPort: socket connect error: " << error.what() << endl;
			return false;
		}

		mConfigured = true;

		// Reset the receive status
		mReceiveResult = -1;

		// start receiving
		StartReceive();

		// send a dummy frame to allow the remote endpoint to know us
		mSocket.send (boost::asio::const_buffers_1 (" ", 1));

		return true;
	}

	// Prepare the frame
	const unsigned char* buff = msg.GetValueBin();
	int len = msg.GetLength() / 8;

	{
		if (len < 2)
			goto too_short;
		
		unsigned char src_len = buff[0];
		unsigned char dst_len = buff[1];

		buff += 2; len -= 2;

		if ((src_len > 8) | (dst_len > 8))
			goto bad_length;

		if (len < src_len + dst_len)
			goto too_short;
	
		char hwsrc[8];
		char hwdst[8];

		memcpy (hwsrc, buff, src_len);
		buff += src_len; len -= src_len;

		memcpy (hwdst, buff, dst_len);
		buff += dst_len; len -= dst_len;

		Bitstring frame = make_ieee_802_15_4_frame (
			hwsrc,	src_len,
			hwdst,	dst_len,
			buff,	len,
			mPanId
		);

		try {
			mSocket.send (boost::asio::const_buffers_1 (frame.GetValueBin(), frame.GetLength()/8));
			return true;
		} catch (boost::system::system_error& error) {
			cerr << "SixLowpanPort: send error: " << error.what() << endl;
			return false;
		}
	}

too_short:
	cerr << "SixLowpanPort::Send() error: frame is too short" << endl;
	return false;

bad_length:
	cerr << "SixLowpanPort::Send() error: invalid address length" << endl;
	return false;
}

