#include "802.15.4-framer.h"
#include <boost/cstdint.hpp>

using namespace std;
using namespace t3devlib;

extern "C"
{
	// @desc Structure representing an IEEE 802.15.4 header
	struct frame_control_field
	{
#if __BYTE_ORDER == __LITTLE_ENDIAN
		uint8_t
			reserved2	:2,
			dam		:2,
			version		:2,
			sam		:2,

			frame_type	:3,
			security_enabled:1,
			frame_pending	:1,
			ack_request	:1,
			intra_pan	:1,
			reserved1	:1;
#elif __BYTE_ORDER == __BIG_ENDIAN
		uint8_t
			sam		:2,
			version		:2,
			dam		:2,
			reserved2	:2,

			reserved1	:1,
			intra_pan	:1,
			ack_request	:1,
			frame_pending	:1,
			security_enabled:1,
			frame_type	:3;
#else
# error "Please fix <bits/endian.h>"
#endif
	};
}

// @desc Sequence number for outgoing frames (incremented each time a new frame
//       is sent)
static uint8_t sSequenceNumber;


// @desc Tells the value of the sam/dam field corresponding to the given HW
//       address length
static uint8_t length2am (uint8_t address_length)
{
	switch (address_length)
	{
	case 2:		return 2;
	case 8:		return 3;
	default:
			return 0;
	}
}

// @desc Tells the HW address length corresponding to the given sam/dam value
static uint8_t am2length (uint8_t address_length)
{
	switch (address_length)
	{
	case 0:
	case 1:
			return 0; // reserved
	
	case 2:		return 2;
	case 3:		return 8;
	}
}

// @desc Append a field to a buffer in the little endian order
// 	 -> the content pointed by src will be added to the buffer in the reverse order
// 	    (last byte first)
static void append_little_endian (Bitstring& buff, const void* src, int len)
{
	for (const uint8_t* p = ((const uint8_t*)src)+len-1 ; p >= src ; p--)
	{
		buff.Append (p, 8);
	}
}

// @desc Copy a field and convert it into the little endian order
// 	 -> the content pointed by src will be copied to the location dst in
// 	    the reverse order (last byte first)
static void memcpy_little_endian (void* dst, const void* src, int len)
{
	uint8_t* q = (uint8_t*) dst;
	for (const uint8_t* p = ((const uint8_t*)src)+len-1 ; p >= src ; p--)
	{
		*q++ = *p;
	}
}

// @desc Convert 16-bit value from host order into network order
static uint16_t my_htons (uint16_t in)
{
	uint16_t out;
	((uint8_t*) &out) [0] = in >> 8;
	((uint8_t*) &out) [1] = in & 0xff;
	return out;
}

// @desc Assemble a IEEE 802.15.4 frame
//
// @param hw_src	source address
// @param hw_src_len	length of the source address
// @param hw_dst	destination address
// @param hw_dst_len	length of the destination address
// @param payload	payload
// @param payload_len	length of the payload
// @param pan_id	PAN ID to be used in this frame
//
// @return		a binary string containing the resulting IEEE 802.15.4
//			frame
Bitstring make_ieee_802_15_4_frame (
	const void* hw_src, int hw_src_len,
	const void* hw_dst, int hw_dst_len,
	const void* payload, int payload_len,
	uint16_t pan_id
)
{
	DBitstring result;

	frame_control_field fcf;

	memset (&fcf, 0, sizeof (fcf));

	fcf.intra_pan		= 1;
	fcf.ack_request		= 0;
	fcf.frame_pending	= 0;
	fcf.security_enabled	= 0;
	fcf.frame_type		= 1; // data frame
	fcf.sam			= length2am (hw_src_len);
	fcf.version		= 1;
	fcf.dam			= length2am (hw_dst_len);

	append_little_endian (result, &fcf, sizeof (fcf));

	result.Append (&sSequenceNumber, 8);
	sSequenceNumber++;

	if (fcf.dam) {
		append_little_endian (result, &pan_id, 2);
		append_little_endian (result, hw_dst, hw_dst_len);
	}
	if (fcf.sam) {
		if (!fcf.dam) {
			append_little_endian (result, &pan_id, 2);
		}
		append_little_endian (result, hw_src, hw_src_len);
	}

	result.Append ((const char*) payload, payload_len*8);

	return result;
error:
	result.Clear();
	return result;
}

// @desc Decode a IEEE 802.15.4 frame
//
// @param frame		pointer to the frame to be decoded
// @param frame_len	length of the frame to be decoded
//
// @param hw_src	buffer that will store the source address
// 			(must be at least 8-byte long)
// @param hw_src_len	will store the length of the source address
//
// @param hw_dst	buffer that will store the destination address
// 			(must be at least 8-byte long)
// @param hw_dst_len	will store the length of the destination address
//
// @param payload	pointer that will point to the payload of the frame
// @param payload_len	will store the length of the payload
//
// @param pan_id	will store the destination PAN ID of this frame
//
// @return		true if successful
//
bool decode_ieee_802_15_4_frame (
	const void* frame, int frame_len,
	void* hw_src, int& hw_src_len,
	void* hw_dst, int& hw_dst_len,
	const void*& payload, int& payload_len,
	uint16_t& pan_id
)
{
	const uint8_t* p = (const uint8_t*) frame;
	const uint8_t* end = p + frame_len;

#define len		(end - p)
#define need_bytes(nb)	if (len < (nb)) return false;

	frame_control_field fcf;
	need_bytes (sizeof (fcf) + 1);
	
	memcpy_little_endian (&fcf, p, sizeof(fcf));
	p += sizeof(fcf) + 1; // ignore seq

	hw_dst_len = am2length (fcf.dam);
	hw_src_len = am2length (fcf.sam);

	if (fcf.dam) {
		need_bytes (2 + hw_dst_len);

		// dst pan id
		memcpy_little_endian (&pan_id, p, 2);
		p += 2;

		// hw dst addr
		memcpy_little_endian (hw_dst, p, hw_dst_len);
		p += hw_dst_len;
	}
	if (fcf.sam) {
		// src pan id
		if (!fcf.dam) {
			// skip the pan id
			need_bytes (2);
			p += 2;
		}

		// src addr
		need_bytes (hw_src_len);
		memcpy_little_endian (hw_src, p, hw_src_len);
		p += hw_src_len;
	}

	payload = p;
	payload_len = len;

	return true;
}


// @desc Reset the IEEE 802.15.4 framer module
//
// This function resets the sequence number used for outgoind IEEE 802.15.4
// frames
void ieee_802_15_4_framer_reset()
{
	sSequenceNumber = 0;
}

